div.error {
  margin: 2em;
  text-align: center;
}

div.ajax-error {
  padding: 1em;
  text-align: center;
  .alert-danger();
}

div.error > h1 {
  font-size: 300%;
  line-height: normal;
}

div.error > p {
  font-size: 200%;
  line-height: normal;
}
